import React, {useCallback} from "react";
import {ColumnFilter, useFilterState} from "../Columns/ColumnFilter";

/**
 * Filter value regex.
 */
const filterRegex = /^([=!><])?([0-9]+)$/;

/**
 * Number column filter.
 */
export const NumberFilter: ColumnFilter<number, NumberFilterState> = {
	filter: (data: number, filterState: NumberFilterState) => {
		// Read filter value.
		const filterValue = filterRegex.exec(filterState.value);
		// No valid filter value, allow everything.
		if (!filterValue?.length) return true;
		// Get current filter modifier and number.
		const filterModifier = filterValue[1];
		const filterNumber = Number.parseFloat(filterValue[2]);
		// Return result based on modifier.
		switch (filterModifier)
		{
			case ">":
				return data > filterNumber;
			case "<":
				return data < filterNumber;
			case "!":
				return data != filterNumber;
			case "=":
			default:
				return data == filterNumber;
		}
	},

	element: <NumberFilterComponent />
};

/**
 * Number filter state.
 */
export interface NumberFilterState
{
	/**
	 * Filter value.
	 */
	value: string;
}

/**
 * Number filter component.
 */
export function NumberFilterComponent()
{
	// Initialize number filter state.
	const [numberFilterState, setNumberFilterState] =
		useFilterState<NumberFilterState>({ value: "" });

	// Handle filter input change.
	const handleChange = useCallback((event: React.ChangeEvent<HTMLInputElement>) => {
		// Save the current filter value.
		setNumberFilterState({
			value: event.currentTarget.value,
		})
	}, [setNumberFilterState]);

	return (
		<input type={"text"} pattern={filterRegex.source} size={1}
		       value={numberFilterState.value} onChange={handleChange} />
	);
}
