import React, {useCallback, useContext, useMemo, useState} from "react";
import {Instance} from "./Instance";
import {ColumnKey, Columns} from "./Column";
import {RowDefinition} from "./Row";
import {Promisable} from "@kernelui/core";
import {SortState, SortType} from "./Sort";

/**
 * Smartable data type.
 */
export type SmartableData<CK extends ColumnKey> = Promisable<(Promisable<RowDefinition<CK>>)[]>;

/**
 * Smartable properties.
 */
export interface SmartableProperties<CK extends ColumnKey>
{
	/**
	 * Table custom class name.
	 */
	className?: string;

	/**
	 * Table data.
	 */
	data: SmartableData<CK>;

	/**
	 * When defined, only show the columns in the array.
	 */
	shownColumns?: CK[];

	/**
	 * Default row element.
	 */
	rowElement?: React.ReactElement;

	/**
	 * Default column heading element.
	 */
	columnHeadingElement?: React.ReactElement;

	/**
	 * Table rows pagination.
	 */
	paginate?: {
		/**
		 * Number of rows per page.
		 */
		pageSize: number;
	};

	/**
	 * Disable sort feature.
	 */
	disableSort?: boolean;

	/**
	 * Disable filter feature.
	 */
	disableFilter?: boolean;
}

/**
 * Smartable main object.
 */
export type Smartable<CK extends ColumnKey> = {
	/**
	 * Table component.
	 */
	Table: React.FunctionComponent<SmartableProperties<CK>>;
};

/**
 * Smartable columns infered type.
 */
export type SmartableColumns<CK> = CK extends Smartable<infer CK> ? CK : never;

/**
 * Define a new Smartable.
 */
export function createSmartable<CK extends ColumnKey>({columns}: {
	/**
	 * Smartable table columns.
	 */
	columns: Columns<CK>;
}): Smartable<CK>
{
	/**
	 * Filter columns, given an array of shown columns.
	 * @param columns Full columns definition.
	 * @param shownColumns Shown columns list.
	 */
	const filterColumns = (columns: Columns<CK>, shownColumns: CK[]) => {
		// Initialize the filtered columns object.
		const filteredColumns = {...columns};

		// Build an associative object of shown columns.
		const shownColumnsMap = Object.fromEntries(shownColumns.map((shownColumn) => [shownColumn, true])) as Record<CK, boolean>;

		for (const key of Object.keys(columns) as CK[])
		{ // For each key, if it is not present in the shown columns
			if (!shownColumnsMap?.[key])
				// Delete all columns not present in the shown columns map.
				delete filteredColumns[key];
		}

		// Return filtered columns.
		return filteredColumns;
	};

	return {
		Table: (props: SmartableProperties<CK>) => {
			// Initialize sort state.
			const [sortState, setSortState] = useState({} as Partial<Record<CK, SortState>>);

			// Set sort state of a specific column.
			const setColumnSortState = useCallback((key: CK, sortType: SortType|null): void => {
				// Copy current sort state.
				let newSortState = {...sortState};

				if (sortType)
					// A new sort type for given column has been set.
					newSortState[key] = {
						// Setting new sort type.
						type: sortType,
						// Keeping current order, or creating a new one (from the current state size).
						order: newSortState?.[key]?.order ?? (Object.keys(sortState).length + 1),
					};
				else if (newSortState[key])
				{ // Sort type for given column has been reset, removing it.
					const removedOrderKey = newSortState[key]?.order;
					delete newSortState[key];

					// Decrement all remaining greater orders by one, as there is now one less sorted column.
					newSortState = Object.fromEntries((Object.entries(newSortState) as [CK, SortState][]).map(
						// For each column sort state...
						([columnKey, {order: columnSortOrder, ...columnSortState}]) => (
							//... copy the current column sort state, just decrement its order.
							[columnKey, {order: columnSortOrder - (removedOrderKey < columnSortOrder ? 1 : 0), ...columnSortState} as SortState]
						)
					)) as Record<CK, SortState>;
				}

				// Set new sort state.
				setSortState(newSortState);
			}, [sortState, setSortState]);

			// Initialize filter states.
			const [filterStates, setFilterStates] = useState<Partial<Record<CK, any>>>({});

			// Set filter state of a specific column.
			const setColumnFilterState = useCallback((key: CK, filterState: any) => {
				setFilterStates(
					{
						// Copy the other filters states.
						...filterStates,
						// Set the filter state for the given column.
						[key]: filterState,
					}
				);
			}, [filterStates, setFilterStates]);

			// Filter columns from the given property.
			const filteredColumns = props.shownColumns ? filterColumns(columns, props.shownColumns) : columns;

			// Initialize table context value.
			const contextValue = useMemo<TableContextData<CK>>(() => ({
				columns: filteredColumns,
				columnsSortState: sortState,
				setColumnSortState: setColumnSortState,
				columnsFilterStates: filterStates,
				setColumnFilterState: setColumnFilterState,
				...props,
			}), [sortState, setSortState, filterStates, setColumnFilterState, filteredColumns, props]);

			return (
				<TableContext.Provider value={contextValue}>
					<Instance columns={filteredColumns} {...props} />
				</TableContext.Provider>
			);
		},

	};
}

/**
 * Smartable table context data.
 */
export interface TableContextData<CK extends ColumnKey> extends SmartableProperties<CK>
{
	/**
	 * Current table columns.
	 */
	columns: Columns<CK>;

	/**
	 * Current table columns sort state.
	 */
	columnsSortState: Partial<Record<CK, SortState>>;

	/**
	 * Set given table column sort state.
	 * @param key The column key for which to set the sort type.
	 * @param sortType The sort type to set for the given column. NULL to reset sort state.
	 */
	setColumnSortState: (key: CK, sortType: SortType|null) => void;

	/**
	 * Current table columsn filter states.
	 */
	columnsFilterStates: Partial<Record<CK, any>>;

	/**
	 * Set given table column filter state.
	 * @param key The column key for which to set the filter state.
	 * @param filterState The filter state to set for the given column.
	 */
	setColumnFilterState: (key: CK, filterState: any) => void;
}

/**
 * Smartable table context.
 */
const TableContext = React.createContext<TableContextData<ColumnKey>>(undefined);

/**
 * Hook to get current Smartable table context.
 * @param smartable The Smartable main object.
 */
export function useTable<CK extends ColumnKey>(smartable?: Smartable<CK>): TableContextData<CK>
{
	return useContext(TableContext) as TableContextData<CK>;
}
